from django.db import models

from mptt.models import MPTTModel, TreeForeignKey

from faq.db_elastic.engine import es
from .validators import (
    tags_in_text_validate,
    image_size_validate,
    image_types_validate,
)

STATUSES = (
    (True, 'Активный'),
    (False, 'Неактивный'),
)


def _upload_to(instance: models.base.ModelBase, filename: str):
    return f'FAQ/{instance._meta.object_name}/images/{filename}'


class FAQArticle(MPTTModel):
    answer_id = models.CharField(verbose_name='ID статьи',
                                 default='-1', max_length=100,
                                 help_text='заполняется автоматически')
    title = models.CharField(verbose_name='Название', max_length=100,
                             validators=[tags_in_text_validate])
    text = models.TextField(verbose_name='Описание', max_length=900,
                            validators=[tags_in_text_validate])
    image = models.ImageField(verbose_name='Картинка описания',
                              upload_to=_upload_to,
                              blank=True, null=True,
                              validators=[image_size_validate,
                                          image_types_validate])
    article = TreeForeignKey('self', verbose_name='Статья',
                             on_delete=models.CASCADE,
                             null=True, blank=True,
                             related_name='attribute')
    transitions = models.IntegerField(verbose_name='Количество переходов',
                                      default=0)
    status = models.BooleanField(verbose_name='Статус',
                                 choices=STATUSES, default=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.old_article_id = self.article.id if self.article else None

    def save(self, *args, **kwargs):
        if self.answer_id == '-1' or (
                self.article and self.article.id != self.old_article_id
                    ):
            if self.article:
                pk_answer = FAQArticle.objects.filter(
                    level=self.article.lft,
                    article=self.article
                ).count()

                if not FAQArticle.objects.filter(
                        level=self.article.lft, article=self.article,
                        article__answer_id=f'{self.article.answer_id}.{pk_answer}'
                ) or pk_answer == 0:
                    self.answer_id = f'{self.article.answer_id}.{pk_answer + 1}'

                else:
                    self.answer_id = f'{self.article.answer_id}.{pk_answer}'

            else:
                count = FAQArticle.objects.filter(level=0).count()

                if FAQArticle.objects.filter(
                        answer_id=str(count)
                ) or count == 0:
                    self.answer_id = str(count + 1)

                else:
                    self.answer_id = str(count)

        super().save()

        es.update_article(instance=self)

    def delete(self, *args, **kwargs):
        es.delete_elem(obj_id=self.id, index='faq_article')
        super().delete()

    def __str__(self):
        return f"{self.answer_id}. {self.title}"

    class MPTTMeta:
        order_insertion_by = ['title']
        parent_attr = 'article'

    class Meta:
        ordering = ['answer_id']
        verbose_name = 'FAQ статья'
        verbose_name_plural = 'FAQ статьи'


class FAQArticleContent(models.Model):
    article = models.ForeignKey(FAQArticle, on_delete=models.CASCADE,
                                related_name='content')
    text = models.TextField(verbose_name='Текст', max_length=1000,
                            null=True, blank=True,
                            validators=[tags_in_text_validate])
    image = models.ImageField(verbose_name='Картинка',
                              upload_to=_upload_to,
                              blank=True, null=True,
                              validators=[image_size_validate,
                                          image_types_validate])

    def __str__(self):
        return f"{self.id} -- {self.article}"

    class Meta:
        verbose_name = 'Контент'
        verbose_name_plural = 'Контент'



messages = {
    'MainMenu': {
        'text_ru': 'Главное меню',
        'buttons': [
            {'reply': True, 'row': 0, 'text_ru': 'FAQ'},
        ]
    },
    'MainFAQ': {
        'text_ru': 'Это FAQ по роботе с клиентмами.\n\n'
                   'Можешь искать с помощью кнопок или по ключевому слову. '
                   'Например "грубость"',
        'buttons': [
            {'callback_data': 'show_article#', 'row': 0, 'text_ru': 'Смотреть статью целиком'},
        ]
    },
    'NoArticles': {
        'text_ru': 'В базе ещё нет ни одной статьи',
    },
    'NothingFound': {
        'text_ru': 'По Вашему запросу ничего не найдено.\nПопробуйте ещё раз',
    },

}



from django.conf import settings

from loguru import logger
from typing import Union, Literal
from elasticsearch import Elasticsearch, exceptions
from elasticsearch_dsl import Q, Search, UpdateByQuery

from .document import FaqArticle, FaqArticleContent


class ElasticEngine(Elasticsearch):
    def __init__(self):
        self.is_active = False

        try:
            super().__init__(hosts=settings.ELASTICSEARCH)

            FaqArticle.init(using=self)
            FaqArticleContent.init(using=self)

            self.is_active = True

        except AttributeError:
            logger.error(
                f"You have not defined the ELASTICSEARCH variable in settings.py"
            )

        except exceptions.ConnectionError as e:
            logger.error(
                f"{e}\n"
                f"____________________________________________________\n"
                f"Maybe you have not started the ElasticSearch service"
                f"\n____________________________________________________"
            )

    def search_faq(self, query: str, status: bool = True):
        """
        :param query: Text for search
        :param status: FAQ status (True = Активный, False = Неактивный)

        :return: [(django_id, score)]
        """

        if self.is_active:
            resp_article = Search(using=self, index='faq_article').query(
                Q('multi_match', query=query, fields=['title', 'text']) &
                Q('match', status=status)
            ).execute()

            resp_content = Search(using=self,
                                  index='faq_article_content').query(
                Q('match', text=query) &
                Q('match', status=status)
            ).execute()

            articles = [(
                sec['_source']['id'],
                sec['_score']
            ) for sec in resp_article['hits']['hits'] if
                sec['_source'].to_dict().get(
                    'id'
                )]

            contents = [(
                sec['_source']['article_id'],
                sec['_score']
            ) for sec in resp_content['hits']['hits'] if
                sec['_source'].to_dict().get(
                    'article_id'
                ) and sec['_source']['article_id'] not in articles[0]]

            return sorted(set(articles + contents), key=lambda hits: hits[1])
        return set()

    def _get_elem(self, obj_id: Union[str, int],
                  index: Literal['faq_article', 'faq_article_content']):
        if self.is_active:
            return Search(using=self, index=index).query("match", id=obj_id).execute()
        return None

    def delete_elem(self, obj_id: Union[str, int],
                    index: Literal['faq_article', 'faq_article_content']):
        if self.is_active:
            Search(using=self, index=index).query("match", id=obj_id).delete()

    def update_article(self, instance):
        if self.is_active:
            elem = self._get_elem(instance.id, index='faq_article')

            if elem:
                UpdateByQuery(
                    using=self,
                    index='faq_article',
                ).query(
                    "match", id=instance.id
                ).script(
                    inline='ctx._source.title = params.local_title; '
                           'ctx._source.text = params.local_text; '
                           'ctx._source.status = params.local_status;',
                    params={
                        "local_title": instance.title,
                        "local_text": instance.text,
                        "local_status": instance.status
                    },
                ).execute()

            else:
                FaqArticle(
                    id=instance.id,
                    title=instance.title,
                    text=instance.text,
                    status=instance.status,
                ).save(using=self)

    def update_content(self, instance):
        if self.is_active:
            elem = self._get_elem(instance.id, index='faq_article_content')

            if elem:
                UpdateByQuery(
                    using=self,
                    index='faq_article_content',
                ).query(
                    "match", id=instance.id
                ).script(
                    inline='ctx._source.text = params.local_text; '
                           'ctx._source.status = params.local_status;',
                    params={
                        "local_text": instance.text,
                        "local_status": instance.article.status
                    },
                ).execute()

            else:
                FaqArticleContent(
                    article_id=instance.article.id,
                    id=instance.id,
                    text=instance.text,
                    status=instance.article.status
                ).save(using=self)


es = ElasticEngine()



from django.db.models.query import QuerySet

from typing import Union
from telebot import types as tp


class DotDict(dict):
    """
     Required for accessing the parameters of the buttons through a dot
        and for the possibility of redefining the values of these parameters
    """

    def __new__(cls, *args, **kwargs):
        self = dict.__new__(cls, *args, **kwargs)
        self.__dict__ = self
        return self


class FAQKeyboards:
    """
     Buttons & markups for FAQ
    """

    def __init__(self):
        self.FAQ_BTN = DotDict(
            text='FAQ',
            callback='faq#'
        )
        self.ARTICLE_BTN = DotDict(
            text='Смотреть статью целиком',
            callback='show_article#'
        )
        self.TREE_BTN = DotDict(
            callback=f'tree_article#'
        )
        self.BACK_BTN = DotDict(
            text='🔙 Назад'
        )

    # ----------------- Reply keyboards -----------------

    def add_faq_reply_btn(self, markup: tp.ReplyKeyboardMarkup) \
            -> tp.ReplyKeyboardMarkup:
        return markup.add(tp.KeyboardButton(text=self.FAQ_BTN.text))

    def faq_reply_keyboard(self, row_width: int = 1) -> tp.ReplyKeyboardMarkup:
        return self.add_faq_reply_btn(markup=tp.ReplyKeyboardMarkup(
            resize_keyboard=True, one_time_keyboard=True, row_width=row_width
        ))

    # ----------------- Inline keyboards -----------------

    def back_btn(self, callback: str, keyboard: bool = False) \
            -> Union[tp.InlineKeyboardButton, tp.InlineKeyboardMarkup]:
        btn = tp.InlineKeyboardButton(
            self.BACK_BTN.text, callback_data=callback
        )

        if keyboard:
            return tp.InlineKeyboardMarkup().add(btn)
        return btn

    def article_btn(self, article_id: Union[int, str], keyboard: bool = False) \
            -> Union[tp.InlineKeyboardButton, tp.InlineKeyboardMarkup]:
        btn = tp.InlineKeyboardButton(
            self.ARTICLE_BTN.text,
            callback_data=f'{self.ARTICLE_BTN.callback}{article_id}'
        )

        if keyboard:
            return tp.InlineKeyboardMarkup().add(btn)
        return btn

    def add_faq_inline_btn(self, markup: tp.InlineKeyboardMarkup) \
            -> tp.InlineKeyboardMarkup:
        return markup.add(
            tp.InlineKeyboardButton(text=self.FAQ_BTN.text,
                                    callback_data=self.FAQ_BTN.callback)
        )

    def faq_inline_markup(self, row_width: int = 1) -> tp.InlineKeyboardMarkup:
        return self.add_faq_inline_btn(markup=tp.InlineKeyboardMarkup(
            row_width=row_width
        ))

    def article_keyboard(self, articles: QuerySet,
                         selected: Union[int, str, None] = None,
                         back: Union[int, str, None] = None) \
            -> tp.InlineKeyboardMarkup:
        markup = tp.InlineKeyboardMarkup(row_width=1)

        if selected:
            markup.add(
                self.article_btn(article_id=selected)
            )

        for article in articles:
            markup.add(
                tp.InlineKeyboardButton(
                    article.title,
                    callback_data=f'{self.TREE_BTN.callback}{article.id}'
                )
            )

        if back is not None:
            markup.add(self.back_btn(
                callback=f'{self.TREE_BTN.callback}{back}')
            )

        return markup



from loguru import logger
from elasticsearch_dsl import Search

from .engine import es


def _match_all_article(index='faq_article'):
    return Search(using=es, index=index).query("match_all")


def _match_content(article_id, index='faq_article_content'):
    return Search(using=es, index=index).query("match", article_id=article_id)


def faq_article_delete():
    res = _match_all_article().delete()
    logger.success(res)


def view_faq_articles():
    for art in _match_all_article().execute()['hits']['hits']:
        logger.debug(art['_source'])

        for cont in _match_content(art['_source']['id']).execute()['hits']['hits']:
            logger.info(cont['_source'])



from django.contrib.postgres.search import TrigramSimilarity

from typing import Optional, Union

from faq.db_elastic.engine import es
from .models import FAQArticle, FAQArticleContent


def _add_article_transition(article: Optional[FAQArticle] = None):
    if article:
        article.transitions += 1
        article.save()


def get_faq_article(article_id: Union[str, int]):
    return FAQArticle.objects.filter(pk=article_id, status=True).first()


def get_faq_article_n_content(article_id: int):
    article = FAQArticle.objects.prefetch_related('content').get(
        id=article_id,
        status=True
    )
    _add_article_transition(article=article)

    return article, article.content.all()


def get_articles_by_level(article_id: Union[str, int, None] = None):
    if article_id:
        return FAQArticle.objects.filter(article_id=article_id, status=True)

    else:
        return FAQArticle.objects.filter(level=0, status=True)


def trigram_search(text: str, score: float = 0.3):
    articles = list(FAQArticle.objects.annotate(
        similarity=TrigramSimilarity('title', text)
                   + TrigramSimilarity('text', text)
    ).filter(similarity__gt=score).order_by('-similarity'))

    contents = [c.article for c in
                FAQArticleContent.objects.annotate(
                    similarity=TrigramSimilarity('text', text)
                ).filter(similarity__gt=score).order_by('-similarity')]

    return articles + contents  # articles have priority


def search_faq_by_text(text: str, answer_count: int = 5):
    answers = es.search_faq(text)                       # ElasticSearch

    if answers:
        articles = []

        for i, _ in answers[:answer_count]:
            if article := get_faq_article(article_id=i):
                articles.append(article)

    else:
        articles = trigram_search(text)[:answer_count]  # Trigram similarity

    return articles



from django.apps import AppConfig


class FaqConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'faq.faq_admin'
    verbose_name = 'FAQ'
    verbose_name_plural = 'FAQ'



# Generated by Django 4.0.2 on 2022-02-18 11:28

from django.contrib.postgres.operations import TrigramExtension
from django.db import migrations


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        TrigramExtension(),  # sets up the extension in PostgreSQL
    ]



from django.db.models import FileField, ImageField

from typing import Optional, Union
from telebot import TeleBot
from telebot import types as tp

from .keyboards import FAQKeyboards
from .translations import messages as ms
from faq.faq_admin import orm_utils as orm
from bot_storage.storage import BaseStorage


class FAQDispatcher(FAQKeyboards):

    def __init__(self, tbot: TeleBot, storage: BaseStorage,
                 faq_state: str = 'faq_search#', answer_count: int = 5):
        """
        :param tbot: Telegram bot instance
        :param storage: Storage instance for user state & user data
        :param faq_state: State for user FAQ search
        :param answer_count: Search response limit
        """
        self.tb = tbot
        self.st = storage
        self.faq_state = faq_state
        self.answer_count = answer_count
        super().__init__()

    def _send_message(self, chat_id: Union[str, int],
                      mess_text: Optional[str] = None,
                      file: Optional[Union[FileField, ImageField]] = None,
                      markup: Union[tp.ReplyKeyboardMarkup,
                                    tp.ReplyKeyboardRemove,
                                    tp.InlineKeyboardMarkup,
                                    None] = None):
        """
         Base func for sending text messages with or without media-files
        """
        if file:
            if file.name.endswith('.mp4'):
                self.tb.send_video(chat_id, video=file, caption=mess_text,
                                   reply_markup=markup)

            elif file.name.endswith('.gif'):
                self.tb.send_animation(chat_id, animation=file, caption=mess_text,
                                       reply_markup=markup)

            elif file.name.endswith('.webp'):
                self.tb.send_sticker(chat_id=chat_id, sticker=file)
                self.tb.send_message(chat_id, mess_text, reply_markup=markup)

            else:
                self.tb.send_photo(chat_id, photo=file, caption=mess_text,
                                   reply_markup=markup)

        else:
            self.tb.send_message(chat_id, mess_text, reply_markup=markup)

    @staticmethod
    def _caption(obj):
        return f'<b>{obj.title}</b>\n\n{obj.text}'

    def send_faq(self, user_id: Union[str, int],
                 article_id: Union[str, int, None] = None):
        """
         Sends message with your articles

        :param user_id: User's Telegram id
        :param article_id: Article id (from Django DB)
        """

        # set state
        self.st.set_user_state(user_id, state=self.faq_state)

        # send articles
        if article_id:
            if article := orm.get_faq_article(article_id=article_id):
                return self._send_message(
                    chat_id=user_id,
                    mess_text=self._caption(article),
                    file=article.image,
                    markup=self.article_keyboard(
                        articles=orm.get_articles_by_level(article_id=article_id),
                        selected=article_id,
                        back=article.article.id if article.article else ''
                    )
                )

        else:
            if articles := orm.get_articles_by_level():
                return self._send_message(
                    chat_id=user_id,
                    mess_text=ms['MainFAQ']['text_ru'],
                    markup=self.article_keyboard(
                        articles=articles,
                        selected=article_id
                    )
                )

        return self._send_message(
            chat_id=user_id,
            mess_text=ms['NoArticles']['text_ru']
        )

    def send_article(self, user_id: Union[str, int],
                     article_id: Union[str, int]):
        """
         Sends message with concrete article

        :param user_id: User's Telegram id
        :param article_id: Article id (from Django DB)
        """
        article, content = orm.get_faq_article_n_content(article_id=article_id)

        if self.st.get_user_data(user_id).get('state') == self.faq_state:
            back_keyboard = self.back_btn('faq_search#', True)
        else:
            back_keyboard = self.back_btn(f'tree_article#{article_id}', True)

        self._send_message(
            user_id, mess_text=self._caption(obj=article), file=article.image,
            markup=None if content else back_keyboard
        )

        last_cnt_num = content.count() - 1

        for idx, obj in enumerate(content):  # TODO: сделать отправку через CELERY
            self._send_message(
                user_id, mess_text=obj.text, file=obj.image,
                markup=back_keyboard if idx == last_cnt_num else None
            )

    def send_faq_search(self, user_id: Union[str, int],
                        text: Optional[str] = None):
        """
         Sends a search response to a user's text query

        :param user_id: User's Telegram id
        :param text: User's input
        """
        if not text:
            text = self.st.get_user_data(user_id).get('input')

        if answers := orm.search_faq_by_text(text, self.answer_count):
            # set data
            self.st.set_user_data(
                user_id, data={'state': self.faq_state, 'input': text}
            )

            # send articles
            for obj in answers:  # TODO: сделать отправку через CELERY
                self._send_message(
                    user_id, mess_text=self._caption(obj=obj), file=obj.image,
                    markup=self.article_btn(article_id=obj.id, keyboard=True)
                )

        else:
            self._send_message(
                chat_id=user_id,
                mess_text=ms['NothingFound']['text_ru']
            )



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="django-bot-faq",
	version="0.2.0",
	author="Ivan Romanchenko",
	author_email="vanvanych789@gmail.com",
	description="FAQ module",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/IvanRomanchenko/django-bot-faq",
	packages=[
		"faq",
		"faq.faq_admin",
		"faq.faq_admin.migrations",
		"faq.db_elastic",
		"faq.faq_tbot"
	],
	include_package_data=True,
	classifiers=[
		"Programming Language :: Python :: 3.10",
		"Framework :: Django :: 4.0",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	python_requires=">=3.6",
	install_requires=[
		"Django==4.0.2",
		"django-mptt==0.13.4",
		"django-cleanup==6.0.0",
		"pyTelegramBotAPI==4.4.0",
		"psycopg2==2.9.3",
		"bot-storage==1.0.1",
		"loguru==0.6.0",
		"elasticsearch==7.17.0",
		"elasticsearch-dsl==7.4.0",
	]
)



from django.db import models
from django.contrib import admin, messages
from django.contrib.admin.widgets import AdminFileWidget
from django.utils.safestring import mark_safe

from .models import FAQArticle, FAQArticleContent
from faq.db_elastic.engine import es


class AdminImageWidget(AdminFileWidget):

    def render(self, name, value, attrs=None, renderer=None):
        output = []

        if value and getattr(value, "url", None):
            image_url = value.url

            output.append(
                f'<a href="{image_url}" target="_blank">'
                f'<img src="{image_url}" alt="{value}" height="200" style="object-fit: cover;"/>'
                f'</a>'

                # Excludes extra fields
                #
                # f'<span class="clearable-file-input">'
                # f'<input type="checkbox" name="image-clear" id="image-clear_id">'
                # f'<label for="image-clear_id">Clear</label>'
                # f'</span><br>Change: '
                # f'<input type="file" name="image" accept="image/*" id="id_image">'
            )

        output.append(super().render(name, value, attrs))

        return mark_safe(u''.join(output))


class FAQArticleContentInline(admin.StackedInline):
    model = FAQArticleContent
    extra = 0
    formfield_overrides = {models.ImageField: {'widget': AdminImageWidget}}


@admin.register(FAQArticle)
class FAQArticleAdmin(admin.ModelAdmin):
    list_display = ('answer_id', 'title', 'get_text',
                    'attribute', 'transitions', 'status')
    list_display_links = ('answer_id', 'title')
    fields = ('get_answer_id', 'title', 'text', 'image',
              'article', 'transitions', 'status')

    list_filter = ('status',)
    search_fields = ('answer_id', 'title', 'text', 'article__title')
    actions = ('status_active', 'status_not_active')

    inlines = [FAQArticleContentInline]

    formfield_overrides = {models.ImageField: {'widget': AdminImageWidget}}

    @admin.display(description='Текст', ordering='text')
    def get_text(self, obj):
        return obj.text[0:50]

    @admin.display(description='Атрибут')
    def attribute(self, obj):
        return mark_safe(
            '<br>'.join(
                x.title for x in FAQArticle.objects.filter(
                    article=obj,
                )
            )
        )

    @admin.display(description='id статьи', ordering='description')
    def get_answer_id(self, obj):
        return 'не указан' if obj.answer_id == '-1' else obj.answer_id

    @admin.display(description='Активный')
    def status_active(self, request, queryset):
        queryset.update(status=True)

    @admin.display(description='Неактивный')
    def status_not_active(self, request, queryset):
        queryset.update(status=False)

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return 'transitions', 'get_answer_id', 'article',
        return 'transitions', 'get_answer_id',

    def delete_queryset(self, request, queryset):
        for el in queryset:
            es.delete_elem(obj_id=el.id, index='faq_article')
        queryset.delete()

    def save_formset(self, request, form, formset, change):
        instances = formset.save(commit=False)

        for obj in formset.deleted_objects:
            es.delete_elem(obj_id=obj.id, index='faq_article_content')
            obj.delete()

        for instance in instances:
            if instance.text or instance.image:
                instance.save()
                es.update_content(instance=instance)

            else:
                messages.add_message(
                    request,
                    messages.ERROR,
                    '"Контент" должен содержать текст и/или картинку'
                )

        formset.save_m2m()

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.related_model \
                and request.resolver_match.kwargs.get('object_id'):
            object_id = request.resolver_match.kwargs['object_id']
            kwargs["queryset"] = db_field.related_model.objects.exclude(
                id=object_id
            )

        return super().formfield_for_foreignkey(db_field, request, **kwargs)



from elasticsearch_dsl import Boolean, Document, Integer, Text


class IndexMixin:
    settings = {
        'number_of_shards': 1,
        'number_of_replicas': 0,
        "analysis": {
            "analyzer": {
                "russian": {
                    "tokenizer": "standard",
                }
            }
        }
    }


class FaqArticle(Document):
    id = Integer()

    answer_id = Text()
    title = Text()
    text = Text()
    status = Boolean()

    class Index(IndexMixin):
        name = 'faq_article'


class FaqArticleContent(Document):
    id = Integer()

    article_id = Integer()
    text = Text()
    status = Boolean()

    class Index(IndexMixin):
        name = 'faq_article_content'



from django.core.exceptions import ValidationError

import re


def image_size_validate(value):
	if value.size > 5_000_000:
		raise ValidationError("Превышен допустимый размер в 5 MB!")


def image_types_validate(value):
	file_type = value.name.split('.')[-1]

	if file_type not in ('png', 'jpg', 'jpeg', 'gif', 'webp'):
		raise ValidationError(
			"Недопустимое расширение файла! "
			"Допускаются только 'png', 'jpg', 'jpeg', 'gif', 'webp'"
		)


def tags_in_text_validate(value):
	open_tags = ('<b', '<strong', '<i', '<em', '<u', '<ins', '<s', '<strike', '<del', '<tg-spoiler', '<code', '<pre', '<a href=')
	close_tags = ('</b>', '</strong>', '</i>', '</em>', '</u>', '</ins>', '</s>', '</strike>', '</del>', '</tg-spoiler>', '</code>', '</pre>', '</a>')
	err = ValidationError(
		"В тексте есть незакрытые теги!"
		"Проверьте правильность написания."
		"Примеры находятся по ссылке https://core.telegram.org/bots/api#html-style"
	)

	if '<' in value or '>' in value:
		if len(re.findall(r'[<>]', value)) % 4 != 0:
			raise err

		else:
			for open_tag, close_tag in zip(open_tags, close_tags):
				if open_tag in value and close_tag not in value or close_tag in value and open_tag not in value:
					raise err
